#define SPHO_ENABLE_DEBUG_PRINT

#include <sys/queue.h>

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "spho/util.h"

#include "msph/tree.h"


struct tree_parse {
	struct msph_token_stream	*s;

	struct msph_token		curr;

	SLIST_HEAD(msph_token_l, msph_token)	head;
};

#define CURR_TOKEN(parse)	(&parse->curr)

#define CTX(parse)		((parse)->s->ctx)

static void	tree_parse_init(struct tree_parse *,
		    struct msph_token_stream *);
static int	tree_parse_next_token(struct tree_parse *);
static void	tree_parse_push_token(struct tree_parse *, struct msph_token *);

static ssize_t	tree_ind_fprint(FILE *, int, struct msph_tree *);

static void	free_the_tree(struct msph_tree *);

static struct msph_tree_unit	*tree_parse_unit(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_body	*tree_parse_body(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_ident	*tree_parse_ident(struct tree_parse *,
				    struct msph_tree *);

static struct msph_tree_dir *tree_parse_tpdef(struct tree_parse *,
				struct msph_tree *);
static struct msph_tree_dir *tree_parse_assert(struct tree_parse *,
				  struct msph_tree *);
static struct msph_tree_dir *tree_parse_nomindecl(
				    struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_dir *tree_parse_membdecl(
				    struct tree_parse *,
				    struct msph_tree *);

static struct msph_tree_tpexpr	*tree_parse_tpexpr(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_true(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_false(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_name(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_applargs(struct tree_parse *,
				    struct msph_tree_ident*,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_trait(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_conj(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_disj(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_impl(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_arrow(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_box(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_forall(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_paren(struct tree_parse *,
				    struct msph_tree *);

struct msph_tree_root *
msph_tree_makeroot(struct msph_ctx *ctx)
{
	struct msph_tree_root	*root;

	if ((root = malloc(sizeof(*root))) == NULL) {
		MSPH_ERR(ctx, MSPH_ERR_SYS);
		return (NULL);
	}
	T(root)->type = MSPH_TREE_ROOT;
	T(root)->parent = NULL;

	root->ctx = ctx;

	STAILQ_INIT(&root->head);

	return (root);
}


/*
 * Expect to read next token, otherwise set error and do something
 * readres: (int) result of read
 * parse: struct tree_parse * to perform read
 * err_info: value passed to MSPH_ERR_INFO
 * on_err: statement to perform on error
 */
#define EXPECT_READ_NEXT(readres, parse, err_info, on_err) \
	do { \
		if (((readres) = tree_parse_next_token(parse)) == -1) \
			on_err; \
		if ((readres) == 0) { \
			MSPH_ERR_INFO(CTX(parse), MSPH_ERR_TREE_EOF, \
			    err_info); \
			on_err; \
		} \
	} while (0)

/*
 * Expect current token type, otherwise set error and do something
 * parse: struct tree_parse * to perform read
 * toktype: token type expected
 * err_info: value passed to MSPH_ERR_INFO
 * on_err: statement to perform on error
 */
#define EXPECT_CURR_TOKEN(parse, toktype, err_info, on_err) \
	do { \
		if (CURR_TOKEN(parse)->type != toktype) { \
			MSPH_ERR_INFO(CTX(parse), MSPH_ERR_TREE_NOMATCH, \
			    err_info); \
			on_err; \
		} \
	} while (0)

int
msph_tree_parse(struct msph_token_stream *s, struct msph_tree_root *root)
{
	struct tree_parse p;
	struct msph_tree_unit *unit;

	if (T(root)->type != MSPH_TREE_ROOT) {
		MSPH_ERR(s->ctx, MSPH_ERR_INVAL);
		return (-1);
	}

	tree_parse_init(&p, s);

	if ((unit = tree_parse_unit(&p, T(root))) == NULL)
		return (-1);

	STAILQ_INSERT_TAIL(&root->head, unit, entries);

	return (0);
}

ssize_t
msph_tree_fprint(FILE *f, struct msph_tree *tree) {
	return (tree_ind_fprint(f, 0, tree));
}

#define MSPH_TREE_INDENT	"    "

__attribute__((format(printf, 3, 4)))
static ssize_t
ind_fprintf(FILE *f, int indent, const char *fmt, ...)
{
	int res;
	ssize_t ret;
	int i;
	va_list ap;

	ret = 0;
	va_start(ap, fmt);

	for (i = 0; i < indent; i++) {
		res = fprintf(f, "%s", MSPH_TREE_INDENT);
		if (res < 0) {
			return ((ssize_t)res);
		}
		ret += (ssize_t)res;
	}

	res = vfprintf(f, fmt, ap);
	if (res < 0) {
		return ((ssize_t)res);
	}
	ret += (ssize_t)res;
	va_end(ap);

	return (ret);
}

static ssize_t
tree_ind_fprint(FILE *f, int indent, struct msph_tree *tree)
{
	ssize_t ret, res;
	struct msph_tree_root *root;
	struct msph_tree_unit *unit;
	struct msph_tree_body *body;
	struct msph_tree_dir *dir;
	struct msph_tree_nomindecl *nomd;
	struct msph_tree_assert *ass;
	struct msph_tree_tpdef *tpdef;
	struct msph_tree_membdecl *membdecl;
	struct msph_tree_conj *conj;
	struct msph_tree_disj *disj;
	struct msph_tree_impl *impl;
	struct msph_tree_arrow *arrow;
	struct msph_tree_box *box;
	struct msph_tree_forall *forall;
	struct msph_tree_appl *appl;
	struct msph_tree_trait *trait;
	struct msph_tree_name *name;
	struct msph_tree_paren *paren;
	struct msph_tree_ident *ident;
	struct msph_tree_tpexpr *tp;

	ret = 0;
	switch (tree->type) {
	case MSPH_TREE_ROOT:
		root = (struct msph_tree_root *)tree;
		if ((res = ind_fprintf(f, indent, "(root\n")) < 0)
			return (res);
		STAILQ_FOREACH(unit, &root->head, entries) {
			if ((res = tree_ind_fprint(f, indent+1, T(unit)))
			    < 0)
				return (res);
			ret += res;
		}
		if ((res = ind_fprintf(f, indent, ")\n")) < 0)
			return (res);
		break;
	case MSPH_TREE_UNIT:
		unit = (struct msph_tree_unit *)tree;
		if ((res = ind_fprintf(f, indent, "(unit:%s\n", unit->name))
		    < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(unit->body))) < 0)
			return (res);
		ret += res;
		if ((res = ind_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_BODY:
		body = (struct msph_tree_body *)tree;
		if ((res = ind_fprintf(f, indent, "(body\n")) < 0)
			return (res);
		ret += res;
		STAILQ_FOREACH(dir, &body->head, entries) {
			if ((res = tree_ind_fprint(f, indent+1, T(dir)))
			    < 0)
				return (res);
			ret += res;
		}
		if ((res = ind_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_NOMINDECL:
		nomd = (struct msph_tree_nomindecl *)tree;
		if ((res = ind_fprintf(f, indent, "(nomindecl:%u,%u\n",
		    TXT(nomd)->pos.line, TXT(nomd)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(nomd->id))) < 0)
			return (res);
		ret += res;
		if ((res = ind_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_ASSERT:
		ass = (struct msph_tree_assert *)tree;
		if ((res = ind_fprintf(f, indent, "(assert:%u,%u\n",
		    TXT(ass)->pos.line, TXT(ass)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(ass->ltp))) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(ass->rtp))) < 0)
			return (res);
		ret += res;
		if ((res = ind_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_TPDEF:
		tpdef = (struct msph_tree_tpdef *)tree;
		if ((res = ind_fprintf(f, indent, "(tpdef:%u,%u\n",
		    TXT(tpdef)->pos.line, TXT(tpdef)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(tpdef->id))) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(tpdef->tp))) < 0)
			return (res);
		ret += res;
		if ((res = ind_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_MEMBDECL:
		membdecl = (struct msph_tree_membdecl *)tree;
		if ((res = ind_fprintf(f, indent, "(membdecl:%u,%u\n",
		    TXT(membdecl)->pos.line, TXT(membdecl)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(membdecl->id))) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(membdecl->tp))) < 0)
			return (res);
		ret += res;
		if ((res = ind_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_CONJ:
		conj = (struct msph_tree_conj *)tree;
		if ((res = ind_fprintf(f, indent, "(conj:%u,%u\n",
		    TXT(conj)->pos.line, TXT(conj)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(conj->ltp))) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(conj->rtp))) < 0)
			return (res);
		ret += res;
		if ((res = ind_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_DISJ:
		disj = (struct msph_tree_disj *)tree;
		if ((res = ind_fprintf(f, indent, "(disj:%u,%u\n",
		    TXT(disj)->pos.line, TXT(disj)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(disj->ltp))) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(disj->rtp))) < 0)
			return (res);
		ret += res;
		if ((res = ind_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_IMPL:
		impl = (struct msph_tree_impl *)tree;
		if ((res = ind_fprintf(f, indent, "(impl:%u,%u\n",
		    TXT(impl)->pos.line, TXT(impl)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(impl->ltp))) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(impl->rtp))) < 0)
			return (res);
		ret += res;
		if ((res = ind_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_ARROW:
		arrow = (struct msph_tree_arrow *)tree;
		if ((res = ind_fprintf(f, indent, "(arrow:%u,%u\n",
		    TXT(arrow)->pos.line, TXT(arrow)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(arrow->ltp))) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(arrow->rtp))) < 0)
			return (res);
		ret += res;
		if ((res = ind_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_BOX:
		box = (struct msph_tree_box *)tree;
		if ((res = ind_fprintf(f, indent, "(box:%u,%u\n",
		    TXT(box)->pos.line, TXT(box)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(box->inner))) < 0)
			return (res);
		ret += res;
		if ((res = ind_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_FORALL:
		forall = (struct msph_tree_forall *)tree;
		if ((res = ind_fprintf(f, indent, "(forall:%u,%u\n",
		    TXT(forall)->pos.line, TXT(forall)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(forall->id))) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(forall->inner))) < 0)
			return (res);
		ret += res;
		if ((res = ind_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_PAREN:
		paren = (struct msph_tree_paren *)tree;
		if ((res = ind_fprintf(f, indent, "(paren:%u,%u\n",
		    TXT(paren)->pos.line, TXT(paren)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(paren->inner))) < 0)
			return (res);
		ret += res;
		if ((res = ind_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_APPL:
		appl = (struct msph_tree_appl *)tree;
		if ((res = ind_fprintf(f, indent, "(appl:%u,%u\n",
		    TXT(appl)->pos.line, TXT(appl)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(appl->id))) < 0)
			return (res);
		ret += res;
		STAILQ_FOREACH(tp, &appl->head, entries) {
			if ((res = tree_ind_fprint(f, indent+1, T(tp))) < 0)
				return (res);
			ret += res;
		}
		if ((res = ind_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_NAME:
		name = (struct msph_tree_name *)tree;
		if ((res = ind_fprintf(f, indent, "(name:%u,%u\n",
		    TXT(name)->pos.line, TXT(name)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(name->id))) < 0)
			return (res);
		ret += res;
		if ((res = ind_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_TRAIT:
		trait = (struct msph_tree_trait *)tree;
		if ((res = ind_fprintf(f, indent, "(trait:%u,%u\n",
		    TXT(trait)->pos.line, TXT(trait)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, T(trait->body))) < 0)
			return (res);
		ret += res;
		if ((res = ind_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_IDENT:
		ident = (struct msph_tree_ident *)tree;
		if ((res = ind_fprintf(f, indent, "(ident:%u,%u:%s)\n",
		    TXT(ident)->pos.line, TXT(ident)->pos.col, ident->str))
		    < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_TRUE:
		if ((res = ind_fprintf(f, indent, "(true:%u,%u)\n",
		    TXT(tree)->pos.line, TXT(tree)->pos.col)) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_FALSE:
		if ((res = ind_fprintf(f, indent, "(false:%u,%u)\n",
		    TXT(tree)->pos.line, TXT(tree)->pos.col)) < 0)
			return (res);
		ret += res;
		break;
	default:
		SPHO_ASSERT(0);
		break;
	}

	return (ret);
}

static void
tree_parse_init(struct tree_parse *p, struct msph_token_stream *s)
{
	p->s = s;

	memset(&p->curr, 0, sizeof(p->curr));

	SLIST_INIT(&p->head);
}

static int
tree_parse_next_token(struct tree_parse *p)
{
	int ret;
	struct msph_token *tok;
	if (! SLIST_EMPTY(&p->head)) {
		tok = SLIST_FIRST(&p->head);
		SLIST_REMOVE_HEAD(&p->head, entries);
		memcpy(&p->curr, tok, sizeof(p->curr));
		free(tok);
		ret = 1;
	} else {
		ret = (int)msph_token_stream_read(&p->curr, 1, p->s);
	}
#ifdef SPHO_ENABLE_DEBUG_PRINT
	if (ret == 1) {
		char buf[256];
		msph_token_str(buf, sizeof(buf), CURR_TOKEN(p));
		SPHO_DEBUG_PRINT("tree_parse_read_token: curr=%s\n",
		    buf);
	}
#endif
	return (ret);
}

struct msph_tree_unit *
tree_parse_unit(struct tree_parse *p, struct msph_tree *parent)
{
	size_t namelen;
	struct msph_tree_unit *unit;

	if ((unit = malloc(sizeof(*unit))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(unit)->type = MSPH_TREE_UNIT;
	T(unit)->parent = parent;
	SCOPE_DECOR_INIT(&unit->dec);
	unit->body = NULL;

	if ((namelen = strlcpy(unit->name, p->s->name, sizeof(unit->name)))
	    >= sizeof(unit->name)) {
		MSPH_ERR(CTX(p), MSPH_ERR_TOOLONG);
		goto err;
	}

	if ((unit->body = tree_parse_body(p, T(unit))) == NULL)
		goto err;

	return (unit);

err:
	free_the_tree(T(unit));

	return (NULL);
}

static struct msph_tree_body *
tree_parse_body(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_dir *dir;
	struct msph_tree_body *body;
	struct msph_token *tok;

	if ((body = malloc(sizeof(*body))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(body)->type = MSPH_TREE_BODY;
	T(body)->parent = parent;
	STAILQ_INIT(&body->head);

	while ((res = tree_parse_next_token(p)) > 0) {
		dir = NULL;
		switch(CURR_TOKEN(p)->type) {
		case TOK_KW_TYPE:
			SPHO_DEBUG_PRINT("parsing tpdef\n");
			dir = tree_parse_tpdef(p, T(body));
			break;
		case TOK_KW_NOMINAL:
			dir = tree_parse_nomindecl(p, T(body));
			break;
		case TOK_KW_MEMBER:
			dir = tree_parse_membdecl(p, T(body));
			break;
		case TOK_KW_ASSERT:
			dir = tree_parse_assert(p, T(body));
			break;
		default:
			if ((tok = msph_token_copy(CTX(p), CURR_TOKEN(p)))
			    == NULL)
				goto err;
			tree_parse_push_token(p, tok);
			goto ret;
		}

		if (dir == NULL)
			goto err;

		STAILQ_INSERT_TAIL(&body->head, dir, entries);
	}

	if (res == -1)
		goto err;
ret:
	return (body);
err:
	free_the_tree(T(body));

	return (NULL);
}

void
tree_parse_push_token(struct tree_parse *p, struct msph_token *tok)
{
	SLIST_INSERT_HEAD(&p->head, tok, entries);
	CURR_TOKEN(p)->type = TOK_INVAL;
}


struct msph_tree_dir *
tree_parse_tpdef(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_tpdef *def;

	EXPECT_CURR_TOKEN(p, TOK_KW_TYPE, MSPH_TREE_TPDEF, return (NULL));

	if ((def = malloc(sizeof(*def))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(def)->type = MSPH_TREE_TPDEF;
	T(def)->parent = parent;
	TXT(def)->pos = CURR_TOKEN(p)->pos;
	// TODO add scope decor when generalizing to type operators

	EXPECT_READ_NEXT(res, p, MSPH_TREE_TPDEF, goto err);
	if ((def->id = tree_parse_ident(p, T(def))) == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_TPDEF, goto err);
	EXPECT_CURR_TOKEN(p, TOK_EQUALS, MSPH_TREE_TPDEF, goto err);

	EXPECT_READ_NEXT(res, p, MSPH_TREE_TPDEF, goto err);
	if ((def->tp = tree_parse_tpexpr(p, T(def))) == NULL)
		goto err;

	return ((struct msph_tree_dir *)def);
err:
	free_the_tree(T(def));
	return (NULL);
}

struct msph_tree_ident *
tree_parse_ident(struct tree_parse *p, struct msph_tree *parent)
{
	struct msph_tree_ident *id;

	EXPECT_CURR_TOKEN(p, TOK_IDENT, MSPH_TREE_IDENT, return (NULL));

	if ((id = malloc(sizeof(*id))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(id)->type = MSPH_TREE_IDENT;
	T(id)->parent = parent;
	TXT(id)->pos = CURR_TOKEN(p)->pos;
	NOM_DECOR_INIT(&id->dec);

	if (SPHO_STRLCPY(id->str, CURR_TOKEN(p)->data.str, sizeof(id->str))
	    >= sizeof(id->str)) {
		MSPH_ERR(CTX(p), MSPH_ERR_TREE_TOOLONG);
		goto err;
	}

	return (id);
err:
	free_the_tree(T(id));
	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_tpexpr(struct tree_parse *p, struct msph_tree *parent)
{
	struct msph_tree_tpexpr *tp;

	switch (CURR_TOKEN(p)->type) {
	case TOK_AMP:
		tp = tree_parse_conj(p, parent);
		break;
	case TOK_PIPE:
		tp = tree_parse_disj(p, parent);
		break;
	case TOK_IMPL:
		tp = tree_parse_impl(p, parent);
		break;
	case TOK_RARROW:
		tp = tree_parse_arrow(p, parent);
		break;
	case TOK_KW_BOX:
		tp = tree_parse_box(p, parent);
		break;
	case TOK_KW_FORALL:
		tp = tree_parse_forall(p, parent);
		break;
	case TOK_IDENT:
		tp = tree_parse_name(p, parent);
		break;
	case TOK_CONST_TRUE:
		tp = tree_parse_true(p, parent);
		break;
	case TOK_CONST_FALSE:
		tp = tree_parse_false(p, parent);
		break;
	case TOK_LPAREN:
		tp = tree_parse_paren(p, parent);
		break;
	case TOK_LBRACE:
		tp = tree_parse_trait(p, parent);
		break;
	default:
		MSPH_ERR_INFO(CTX(p), MSPH_ERR_TREE_NOMATCH, MSPH_TREE_TPEXPR);
		tp = NULL;
	}

	SPHO_DEBUG_PRINT("returning tp type=%x\n", T(tp)->type);
	return (tp);
}

struct msph_tree_tpexpr *
tree_parse_conj(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_conj *conj;

	EXPECT_CURR_TOKEN(p, TOK_AMP, MSPH_TREE_CONJ, return (NULL));

	if ((conj = malloc(sizeof(*conj))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(conj)->type = MSPH_TREE_CONJ;
	T(conj)->parent = parent;
	TXT(conj)->pos = CURR_TOKEN(p)->pos;
	TP_DECOR_INIT(&conj->htpe.dec);
	conj->ltp = NULL;
	conj->rtp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_CONJ, goto err);
	if ((conj->ltp = tree_parse_tpexpr(p, T(conj))) 
	    == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_CONJ, goto err);
	if ((conj->rtp = tree_parse_tpexpr(p, T(conj)))
	    == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)conj);
err:
	free_the_tree((struct msph_tree *)conj);

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_disj(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_disj *disj;

	EXPECT_CURR_TOKEN(p, TOK_PIPE, MSPH_TREE_DISJ, return (NULL));

	if ((disj = malloc(sizeof(*disj))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(disj)->type = MSPH_TREE_DISJ;
	T(disj)->parent = parent;
	TXT(disj)->pos = CURR_TOKEN(p)->pos;
	TP_DECOR_INIT(&disj->htpe.dec);
	disj->ltp = NULL;
	disj->rtp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_DISJ, goto err);
	if ((disj->ltp = tree_parse_tpexpr(p, (struct msph_tree *)disj)) 
	    == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_DISJ, goto err);
	if ((disj->rtp = tree_parse_tpexpr(p, (struct msph_tree *)disj))
	    == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)disj);
err:
	free_the_tree((struct msph_tree *)disj);

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_impl(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_impl *impl;

	EXPECT_CURR_TOKEN(p, TOK_IMPL, MSPH_TREE_IMPL, return (NULL));

	if ((impl = malloc(sizeof(*impl))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(impl)->type = MSPH_TREE_IMPL;
	T(impl)->parent = parent;
	TXT(impl)->pos = CURR_TOKEN(p)->pos;
	TP_DECOR_INIT(&impl->htpe.dec);
	impl->ltp = NULL;
	impl->rtp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_IMPL, goto err);
	if ((impl->ltp = tree_parse_tpexpr(p, (struct msph_tree *)impl)) 
	    == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_IMPL, goto err);
	if ((impl->rtp = tree_parse_tpexpr(p, (struct msph_tree *)impl))
	    == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)impl);
err:
	free_the_tree((struct msph_tree *)impl);

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_arrow(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_arrow *arrow;

	EXPECT_CURR_TOKEN(p, TOK_RARROW, MSPH_TREE_ARROW, return (NULL));

	if ((arrow = malloc(sizeof(*arrow))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(arrow)->type = MSPH_TREE_ARROW;
	T(arrow)->parent = parent;
	TXT(arrow)->pos = CURR_TOKEN(p)->pos;
	TP_DECOR_INIT(&arrow->htpe.dec);
	arrow->ltp = NULL;
	arrow->rtp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_ARROW, goto err);
	if ((arrow->ltp = tree_parse_tpexpr(p, (struct msph_tree *)arrow)) 
	    == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_ARROW, goto err);
	if ((arrow->rtp = tree_parse_tpexpr(p, (struct msph_tree *)arrow))
	    == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)arrow);
err:
	free_the_tree((struct msph_tree *)arrow);

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_box(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_box *box;

	EXPECT_CURR_TOKEN(p, TOK_KW_BOX, MSPH_TREE_BOX, return (NULL));

	if ((box = malloc(sizeof(*box))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(box)->type = MSPH_TREE_BOX;
	T(box)->parent = parent;
	TXT(box)->pos = CURR_TOKEN(p)->pos;
	TP_DECOR_INIT(&box->htpe.dec);
	box->inner = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_BOX, goto err);
	if ((box->inner = tree_parse_tpexpr(p, T(box))) == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)box);
err:
	free_the_tree(T(box));
	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_forall(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_forall *fa;

	EXPECT_CURR_TOKEN(p, TOK_KW_FORALL, MSPH_TREE_FORALL, return (NULL));

	if ((fa = malloc(sizeof(*fa))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(fa)->type = MSPH_TREE_FORALL;
	T(fa)->parent = parent;
	TXT(fa)->pos = CURR_TOKEN(p)->pos;
	TP_DECOR_INIT(&fa->htpe.dec);
	SCOPE_DECOR_INIT(&fa->dec);
	fa->id = NULL;
	fa->inner = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_FORALL, goto err);
	if ((fa->id = tree_parse_ident(p, T(fa))) == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_FORALL, goto err);
	EXPECT_CURR_TOKEN(p, TOK_DOT, MSPH_TREE_FORALL, goto err);

	EXPECT_READ_NEXT(res, p, MSPH_TREE_FORALL, goto err);
	if ((fa->inner = tree_parse_tpexpr(p, T(fa))) == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)fa);
err:
	free_the_tree(T(fa));
	return (NULL);
}

/* parse name or appl */
struct msph_tree_tpexpr *
tree_parse_name(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_ident *id;
	struct msph_tree_name *name;
	struct msph_tree_tpexpr *ret;
	struct msph_token *tok;

	ret = NULL;
	name = NULL;

	if ((id = tree_parse_ident(p, NULL)) == NULL)
		return (NULL);
	SPHO_DEBUG_PRINT("parsed ident=%s\n", id->str);

	if ((res = tree_parse_next_token(p)) == 1) {
		if (CURR_TOKEN(p)->type == TOK_LBRAK) {
			if ((ret = tree_parse_applargs(p, id, parent)) == NULL)
				goto err;
			return (ret);
		}

		if ((tok = msph_token_copy(CTX(p), CURR_TOKEN(p)))
		    == NULL) {
			goto err;
		}
		tree_parse_push_token(p, tok);
	} else if (res == -1) {
		goto err;
	}

	SPHO_DEBUG_PRINT("ident is name\n");
	if ((name = malloc(sizeof(*name))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		goto err;
	}
	T(name)->type = MSPH_TREE_NAME;
	T(name)->parent = parent;

	TXT(name)->pos = TXT(id)->pos;
	TP_DECOR_INIT(&name->htpe.dec);
	T(id)->parent = T(name);
	name->id = id;

	return ((struct msph_tree_tpexpr *)name);
err:
	free_the_tree(T(id));

	return (NULL);
}

/**
 * Parse argument list part [...], of type application A[...].
 *
 * @param p Parse state
 * @param id Parsed ident A (on failure, this is not freed)
 * @param parent Parent of resulting appl node
 *
 * @return tpexpr node representing A[...].
 */
struct msph_tree_tpexpr *
tree_parse_applargs(struct tree_parse *p, struct msph_tree_ident *id,
    struct msph_tree *parent)
{
	int res;
	struct msph_tree_appl *appl;
	struct msph_tree_tpexpr *tp;

	tp = NULL;

	EXPECT_CURR_TOKEN(p, TOK_LBRAK, MSPH_TREE_APPL, return (NULL));

	if ((appl = malloc(sizeof(*appl))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(appl)->type = MSPH_TREE_APPL;
	T(appl)->parent = parent;
	TP_DECOR_INIT(&appl->htpe.dec);
	appl->id = NULL;
	STAILQ_INIT(&appl->head);

	/* parse the argument list */
	while ((res = tree_parse_next_token(p)) == 1) {
		/* if ']' we have reached end of arguments */
		if (CURR_TOKEN(p)->type == TOK_RBRAK)
			break;

		/* if not start of argument list, check for comma and step
		 * tokenizer */
		if (tp != NULL) {
			EXPECT_CURR_TOKEN(p, TOK_COMMA, MSPH_TREE_APPL,
			    goto err);
			EXPECT_READ_NEXT(res, p, MSPH_TREE_APPL, goto err);
		}

		if ((tp = tree_parse_tpexpr(p, T(appl))) == NULL)
			goto err;

		STAILQ_INSERT_TAIL(&appl->head, tp, entries);
	}

	/* system error */
	if (res == -1)
		goto err;

	/* did not read final ']' because of EOF */
	if (res == 0) {
		MSPH_ERR_INFO(CTX(p), MSPH_ERR_TREE_EOF, MSPH_TREE_APPL);
		goto err;
	}

	/* finally, appl takes ownership of id */
	T(id)->parent = T(appl);
	appl->id = id;
	TXT(appl)->pos = TXT(id)->pos;

	return ((struct msph_tree_tpexpr *)appl);
err:
	free_the_tree(T(appl));

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_true(struct tree_parse *p, struct msph_tree *parent)
{
	struct msph_tree_true *t;

	EXPECT_CURR_TOKEN(p, TOK_CONST_TRUE, MSPH_TREE_TRUE, return (NULL));

	if ((t = malloc(sizeof(*t))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(t)->type = MSPH_TREE_TRUE;
	T(t)->parent = parent;
	TXT(t)->pos = CURR_TOKEN(p)->pos;
	TP_DECOR_INIT(&t->htpe.dec);

	return ((struct msph_tree_tpexpr *)t);
}

struct msph_tree_tpexpr *
tree_parse_false(struct tree_parse *p, struct msph_tree *parent)
{
	struct msph_tree_true *f;

	EXPECT_CURR_TOKEN(p, TOK_CONST_FALSE, MSPH_TREE_FALSE, return (NULL));

	if ((f = malloc(sizeof(*f))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(f)->type = MSPH_TREE_FALSE;
	T(f)->parent = parent;
	TXT(f)->pos = CURR_TOKEN(p)->pos;
	TP_DECOR_INIT(&f->htpe.dec);

	return ((struct msph_tree_tpexpr *)f);
}

struct msph_tree_tpexpr *
tree_parse_paren(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_paren *par;

	EXPECT_CURR_TOKEN(p, TOK_LPAREN, MSPH_TREE_FALSE, return (NULL));

	if ((par = malloc(sizeof(*par))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(par)->type = MSPH_TREE_PAREN;
	T(par)->parent = parent;
	TXT(par)->pos = CURR_TOKEN(p)->pos;
	TP_DECOR_INIT(&par->htpe.dec);
	par->inner = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_PAREN, goto err);

	if ((par->inner = tree_parse_tpexpr(p, T(par))) == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_PAREN, goto err);
	EXPECT_CURR_TOKEN(p, TOK_RPAREN, MSPH_TREE_PAREN, goto err);

	return ((struct msph_tree_tpexpr *)par);
err:
	free_the_tree(T(par));

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_trait(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_trait *trait;

	EXPECT_CURR_TOKEN(p, TOK_LBRACE, MSPH_TREE_TRAIT, return (NULL));

	if ((trait = malloc(sizeof(*trait))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(trait)->type = MSPH_TREE_TRAIT;
	T(trait)->parent = parent;
	TXT(trait)->pos = CURR_TOKEN(p)->pos;
	TP_DECOR_INIT(&trait->htpe.dec);
	SCOPE_DECOR_INIT(&trait->dec);
	trait->body = NULL;

	if ((trait->body = tree_parse_body(p, T(trait))) == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_TRAIT, goto err);
	EXPECT_CURR_TOKEN(p, TOK_RBRACE, MSPH_TREE_TRAIT, goto err);

	return ((struct msph_tree_tpexpr *)trait);
err:
	free_the_tree(T(trait));

	return (NULL);
}

struct msph_tree_dir *
tree_parse_nomindecl(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_nomindecl *nomd;

	EXPECT_CURR_TOKEN(p, TOK_KW_NOMINAL, MSPH_TREE_NOMINDECL,
	    return (NULL));

	if ((nomd = malloc(sizeof(*nomd))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(nomd)->type = MSPH_TREE_NOMINDECL;
	T(nomd)->parent = parent;
	TXT(nomd)->pos = CURR_TOKEN(p)->pos;
	nomd->id = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_NOMINDECL, goto err);
	if ((nomd->id = tree_parse_ident(p, T(nomd))) == NULL)
		goto err;

	return ((struct msph_tree_dir *)nomd);
err:
	free_the_tree(T(nomd));

	return (NULL);
}

struct msph_tree_dir *
tree_parse_membdecl(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_membdecl *membd;

	EXPECT_CURR_TOKEN(p, TOK_KW_MEMBER, MSPH_TREE_MEMBDECL,
	    return (NULL));

	if ((membd = malloc(sizeof(*membd))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(membd)->type = MSPH_TREE_MEMBDECL;
	T(membd)->parent = parent;
	TXT(membd)->pos = CURR_TOKEN(p)->pos;
	membd->id = NULL;
	membd->tp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_MEMBDECL, goto err);
	if ((membd->id = tree_parse_ident(p, T(membd))) == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_MEMBDECL, goto err);
	EXPECT_CURR_TOKEN(p, TOK_COLON, MSPH_TREE_MEMBDECL,
	    goto err);
	EXPECT_READ_NEXT(res, p, MSPH_TREE_MEMBDECL, goto err);

	if ((membd->tp = tree_parse_tpexpr(p, T(membd))) == NULL)
		goto err;

	return ((struct msph_tree_dir *)membd);
err:
	free_the_tree(T(membd));

	return (NULL);
}

struct msph_tree_dir *
tree_parse_assert(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_assert *ass;

	EXPECT_CURR_TOKEN(p, TOK_KW_ASSERT, MSPH_TREE_ASSERT,
	    return (NULL));

	if ((ass = malloc(sizeof(*ass))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(ass)->type = MSPH_TREE_ASSERT;
	T(ass)->parent = parent;
	TXT(ass)->pos = CURR_TOKEN(p)->pos;
	ass->ltp = NULL;
	ass->rtp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_ASSERT, goto err);
	if ((ass->ltp = tree_parse_tpexpr(p, T(ass))) == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_ASSERT, goto err);
	EXPECT_CURR_TOKEN(p, TOK_SUB, MSPH_TREE_ASSERT,
	    goto err);
	EXPECT_READ_NEXT(res, p, MSPH_TREE_ASSERT, goto err);

	if ((ass->rtp = tree_parse_tpexpr(p, T(ass))) == NULL)
		goto err;

	return ((struct msph_tree_dir *)ass);
err:
	free_the_tree(T(ass));

	return (NULL);
}

static void
free_the_tree(struct msph_tree *tree)
{
	/* will anyone ever free the tree? :´( */
	SPHO_POSTCOND((void *)tree != (void *)free);
	// SPHO_POSTCOND(0);
}

